# -*- coding: utf-8 -*-
from xml.etree.ElementTree import Element, SubElement, tostring


def to_string(xml, encoding='utf-8', xml_declaration=False):
    """Outputs an xml.etree.ElementTree.Element object to a string.

    Usage:
        xml = from_dict(data)
        xml_to_string(xml, xml_declaration=True)  # <?xml version ....

    Args:
        encoding (string): the encoding used for the xml.
        xml_declaration (boolean): whether or not to include the xml declaration.
    """
    declaration = '<?xml version="1.0" encoding="{encoding}" ?>'.format(
        encoding=encoding)
    string = tostring(xml).decode(encoding)
    return '{0}{1}'.format(declaration if xml_declaration else '', string)


def from_dict(obj, node_name='root'):
    """Converts a simple dictionary into an XML document.

    Example:

    .. code-block:: python

        data = {
            'test': {
                'nodes': {
                    'node': [
                        'Testing',
                        'Another node'
                    ]
                },
            }
        }
        xml = from_dict(data)  # <test><nodes><node>Testing</node><node>Another node</node></nodes></test>

    Args:
        node_name (string): the initial node name in case there are multiple
                            top level elements.
    """
    def __dict_to_xml(obj, node_name=None, parent_element=None):
        # internal processing for from_dict
        if not isinstance(parent_element, Element):
            if isinstance(obj, dict) and len(obj) == 1:
                node_name, obj = obj.popitem()
            parent_element = Element(node_name)
        if isinstance(obj, dict):
            for key, value in obj.items():
                if isinstance(value, (list, tuple)):
                    __dict_to_xml(value, key, parent_element)
                else:
                    __dict_to_xml(value, key, SubElement(parent_element, key))
        elif isinstance(obj, (list, tuple)):
            for value in obj:
                sub_element = SubElement(parent_element, node_name)
                __dict_to_xml(value, node_name, sub_element)
        else:
            parent_element.text = obj
        return parent_element

    return __dict_to_xml(obj, node_name)



# -*- coding: utf-8 -*-
from copy import deepcopy, copy
from types import ModuleType


def dict_deep_update(d1, d2):

    """Recursively merge two dictionaries.

    Merges two dictionaries together rather than a shallow update().

    Args:
        d1 (dict): The original dict.
        d2 (dict): The dict to merge with d1.

    Returns:
        dict: A new dict containing the merged dicts.
    """
    if not isinstance(d2, dict):
        return d2
    try:
        result = deepcopy(d1)
    except:
        result = copy(d1)
    for k, v in d2.items():
        if k in result and isinstance(result[k], dict):
            result[k] = dict_deep_update(result[k], v)
        elif k in result and isinstance(result[k], (list, tuple)) and isinstance(v, (list, tuple)):
            result[k] = result[k] + v
        else:
            try:
                result[k] = deepcopy(v)
            except:
                result[k] = copy(v)
    return result


def module_to_dict(module, ignore_starts_with=''):

    """Load the contents of a module into a dict.

    Args:
        ignore_starts_with (string): Ignore all module keys that begin with this value.

    Returns:
        dict: The contents of the module as a dict

    Example:

    .. code-block:: python

        # my_module.py contents:
        # variable = 'value'
        import my_module
        a_dict = module_to_dict(my_module)
        a_dict['variable']
    """
    new_dict = {}
    for k in dir(module):
        item = getattr(module, k)
        if not isinstance(item, ModuleType) and not k.startswith(ignore_starts_with):
            new_dict[k] = item
    return new_dict


class MultiDict(dict):

    """A dictionary type that can contain multiple items for a single key.

    Dictionary type that will create a list of values if more than one item is
    set for that particular key.

    Example:

    .. code-block:: python

        multi_dict = MultiDict()
        multi_dict['one'] = 1
        multi_dict['one'] = 'itchi'
        print(multi_dict)  # {'one': [1, 'itchi']}
    """

    def set(self, key, value, replace=False):
        """Add a new item to the dictionary.

        Set the key to value on the dictionary, converting the existing value
        to a list if it is a string, otherwise append the value.

        Args:
            key (string): The key used to the store the value.
            value (mixed): The value to store.
            replace (bool): Whether or not the value should be replaced.

        Example:

        .. code-block:: python

            multi_dict = MultiDict()
            multi_dict.set('item', 'value')  # or multi_dict['item'] = 'value'
        """
        self.__setitem__(key, value, replace)

    def __setitem__(self, key, value, replace=False):
        # See MultiDict.set for more information.
        if replace and key in self:
            del self[key]
        if key in self:
            try:
                self[key].append(value)
            except:
                existing = [self[key]]
                existing.append(value)
                value = existing
        dict.__setitem__(self, key, value)

    def update(self, a_dict):
        for key, value in a_dict.items():
            self[key] = value

    def copy(self):
        return self.__class__(self)

    def deepcopy(self, memo=None):
        return self.__class__(deepcopy(self._flattened(), memo))

    def _flattened(self):
        return dict(((key, value) for key, value in self.items()))

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo=None):
        return self.deepcopy(memo=memo)


class ImmutableMixin:
    _mutable = True

    def _is_immutable(self):
        if not self._mutable:
            raise TypeError('{0} is immutable'.format(self.__class__.__name__))

    def make_immutable(self):
        self._mutable = False


class ImmutableDict(dict, ImmutableMixin):

    """Creates an immutable dict.

    While not truly immutable (_mutable can be changed), it works effectively
    in the same fashion.
    """

    def __init__(self, *args):
        super(ImmutableDict, self).__init__(*args)
        self.make_immutable()

    def __setitem__(self, key, value, replace=False):
        self._is_immutable()  # pragma: no cover
        super(
            ImmutableDict,
            self).__setitem__(key,
                              value,
                              replace)  # pragma: no cover

    def __delitem__(self, key):
        self._is_immutable()

    def __copy__(self):
        duplicate = {}
        for key, value in self.items():
            duplicate[key] = value
        return duplicate

    def __deepcopy__(self, clone):
        duplicate = {}
        for key, value in self.items():
            duplicate[deepcopy(key, clone)] = deepcopy(value, clone)
        return duplicate

    def appendlist(self, key, value):
        self._is_immutable()  # pragma: no cover
        super(ImmutableDict, self).appendlist(key, value)  # pragma: no cover

    def clear(self):
        self._is_immutable()
        super(ImmutableDict, self).clear()  # pragma: no cover

    def copy(self):
        return self.__deepcopy__({})

    def pop(self, key, *args):
        self._is_immutable()  # pragma: no cover
        return super(ImmutableDict, self).pop(key, *args)  # pragma: no cover

    def popitem(self):
        self._is_immutable()  # pragma: no cover
        return super(ImmutableDict, self).popitem()  # pragma: no cover

    def setdefault(self, key, default=None):
        self._is_immutable()  # pragma: no cover
        return (
            super(
                ImmutableDict,
                self).setdefault(key,
                                 default)  # pragma: no cover
        )

    def update(self, *args):
        self._is_immutable()  # pragma: no cover
        super(ImmutableDict, self).update(*args)  # pragma: no cover


class ImmutableMultiDict(MultiDict, ImmutableMixin):

    """Creates an immuatable MultiDict.
    """

    def __init__(self, *args):
        super(ImmutableMultiDict, self).__init__(*args)
        self.make_immutable()

    def __setitem__(self, key, value, replace=False):
        self._is_immutable()  # pragma: no cover
        super(
            ImmutableMultiDict,
            self).__setitem__(key,
                              value,
                              replace)  # pragma: no cover

    def __delitem__(self, key):
        self._is_immutable()  # pragma: no cover

    def appendlist(self, key, value):
        self._is_immutable()  # pragma: no cover
        super(
            ImmutableMultiDict,
            self).appendlist(key,
                             value)  # pragma: no cover

    def clear(self):
        self._is_immutable()  # pragma: no cover
        super(MultiDict, self).clear()  # pragma: no cover

    def pop(self, key, *args):
        self._is_immutable()  # pragma: no cover
        return (
            super(ImmutableMultiDict, self).pop(key, *args)  # pragma: no cover
        )

    def popitem(self):
        self._is_immutable()  # pragma: no cover
        return super(ImmutableMultiDict, self).popitem()  # pragma: no cover

    def setdefault(self, key, default=None):
        self._is_immutable()  # pragma: no cover
        return (
            super(
                ImmutableMultiDict,
                self).setdefault(key,
                                 default)  # pragma: no cover
        )

    def update(self, *args):
        self._is_immutable()  # pragma: no cover
        super(ImmutableMultiDict, self).update(*args)  # pragma: no cover



# -*- coding: utf-8 -*-
from functools import update_wrapper


def cached_property(func):

    """Allows expensive property calls to be cached.

    Example:

    .. code-block:: python

        class MyClass(object):
            @cached_property
            def expensive_call(self):
                # do something expensive

        klass = MyClass()
        klass.expensive_call  # initial call is made
        klass.expensive_call  # return value is retrieved from an internal cache
    """
    prop = '_{name}'.format(name=func.__name__)

    def _get_property(self):
        try:
            value = getattr(self, prop)
        except AttributeError:
            value = func(self)
            setattr(self, prop, value)
        return value

    update_wrapper(_get_property, func)

    def _del_property(self):
        delattr(self, prop)

    return property(_get_property, None, _del_property)



# -*- coding: utf-8 -*-
from contextlib import contextmanager


@contextmanager
def ignored(*exceptions):
    """Provides the ability to not have to write try/catch blocks when just
    passing on the except.

    Thanks to Raymond Hettinger from "Transforming Code into Beautiful
    Idiotmatic Python"
    This will be included in the standard library in 3.4.

    Args:
        exceptions: A list of exceptions to ignore

    Example:

    .. code-block:: python

        # instead of...
        try:
            do_something()
        except:
            pass

        # use this:
        with ignored(Exception):
            do_something()
    """
    try:
        yield
    except exceptions:
        pass



# -*- coding: utf-8 -*-
import os
from setuptools import setup, Command, find_packages
import watson.common


class BaseCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


class PyTest(BaseCommand):
    """Runs all the unit tests.
    """
    def run(self):
        os.system('py.test')
        clean()


class PyPiPublish(BaseCommand):
    """Publishes the code to PyPi.
    """
    def run(self):
        if (confirm('Are you sure you want to push to PyPi?')):
            os.system('py.test')
            os.system('python setup.py sdist upload')
            clean()


def clean():
    """Cleans the source directory of all non-source files.
    """
    os.system('rm -rf .coverage')
    os.system('find . -name "__pycache__" -print0|xargs -0 rm -rf')
    os.system('find . -name "*.egg-info" -print0|xargs -0 rm -rf')
    os.system('rm -rf dist')


def confirm(prompt):
    prompt += ' (Y/N) [n]: '
    while True:
        answer = input(prompt).upper()
        if not answer:
            answer = 'N'
        return False if answer != 'Y' else True

path = os.path.dirname(os.path.abspath(__file__))

with open(os.path.join(path, 'LICENSE')) as f:
    license = f.read()

with open(os.path.join(path, 'README.rst')) as f:
    readme = f.read()

with open(os.path.join(path, 'requirements.txt')) as f:
    requirements = f.read().splitlines()

with open(os.path.join(path, 'requirements-test.txt')) as f:
    test_requirements = f.read().splitlines()


setup(
    name='watson-common',
    version=watson.common.__version__,
    url='http://github.com/watsonpy/watson-common',
    description='Useful common utility functions and classes.',
    long_description=readme,

    author='Simon Coulton',
    author_email='simon at bespohk.com',

    license=license,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Server'
    ],
    platforms=['Python 3.3'],
    keywords=['watson',
              'python3',
              'web framework',
              'framework',
              'wsgi',
              'web',
              'data structures',
              'dict',
              'context managers'],

    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,

    zip_safe=False,
    install_requires=requirements,
    extras_require={
        'test': test_requirements
    },

    cmdclass={
        'test': PyTest,
        'publish': PyPiPublish
    }
)
